/* * 
 * This file is part of bufferpool
 *
 * Copyright (C) 2007 by LScube team <team@streaming.polito.it>
 * See AUTHORS for more details
 * 
 * bufferpool is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * bufferpool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with bufferpool; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA 
 *
 */




#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/mman.h>

#include <bufferpool.h>


/*!@brief This function adds shared memory page of slots.
 * The function remaps the shared memory according to the new size, but
 * it does NOT link new page to previous existing queue:
 * This is done in bp_addpage.
 * WARNING: the function assumes that the caller (bp_write or bp_getslot) locked the buffer mutex
 * @return the first BPSlot of new added page of slots.
 * */
BPSlot *bp_shm_addpage(BPBuffer * buffer)
{
    BPSlot *added;
    BPSlot *slots;
    unsigned int i;
    char *shm_file_name;
    int fd;
    struct stat fdstat;

    // *** slots mapping in shared memory ***
    if (!
        (shm_file_name =
         bp_ipc_name(buffer->filename, BPBUFF_SHM_SLOTSNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR, 0);
    free(shm_file_name);
    if ((fd < 0)) {
        bp_log(FNC_LOG_ERR,
            "Could not open POSIX shared memory (BPSlots): is Felix running?\n");
        return NULL;
    }
    if ((fstat(fd, &fdstat) < 0)) {
        bp_log(FNC_LOG_ERR, "Could not stat %s\n",
            BPBUFF_SHM_SLOTSNAME);
        close(fd);
        return NULL;
    }

    if (((size_t) fdstat.st_size !=
         buffer->control->nslots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
            "Strange size for shared memory! (not the number of slots reported in control struct)\n");
        close(fd);
        return NULL;
    }
    if (ftruncate
        (fd,
         (buffer->control->nslots + BPBUFF_SHM_PAGE) * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
            "Could not set correct size for shared memory object (BPControl)\n");
        close(fd);
        return NULL;
    }
    if (munmap(buffer->slots, buffer->known_slots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR, "Could not unmap previous slots!!!\n");
        close(fd);
        return NULL;
    }
    slots =
        mmap(NULL,
         (buffer->control->nslots + BPBUFF_SHM_PAGE) * sizeof(BPSlot),
         PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    close(fd);
    if (slots == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        return NULL;
    }
    // inizialization of BPSlots added
    for (i = buffer->control->nslots;
         i < buffer->control->nslots + BPBUFF_SHM_PAGE - 1; i++) {
        slots[i].refs = 0;
        slots[i].slot_seq = 0;
        slots[i].next = i + 1;
    }
    slots[i].refs = 0;
    slots[i].slot_seq = 0;
    slots[i].next = -1;    // last added slot in shm new page has next slot to NULL: bp_addpage will link it correctly.

    added = &slots[buffer->control->nslots];
    buffer->slots = slots;
    buffer->control->nslots += BPBUFF_SHM_PAGE;
    buffer->known_slots = buffer->control->nslots;

    return added;
}
